%{
/* C declarations */

#define YYDEBUG 1
#define YYERROR_VERBOSE 1

%}
/* BISON declarations */

/* Operator precedence - least precedence group */
%left IMP EQV   /* 1101, 1001 */
%left XOR OR    /* 0110, 0111 */
%left AND       /* 0001 */

%right NOT  /* 0=1 1=0 */

/* Operator precedence - middle precedence group */
%left '<' '>' '=' LESSOREQUAL GREATEROREQUAL NOTEQUAL

/* Operator precedence - top precedence group */
%left '+' '-'
%left MOD
%left '\\'
%left '*' '/'

%right NEGATIONPRIORITY
%right '^'

%token CRLF COLONCHAR

%token IF THEN ELSE ELSEIF ENDIF

%token SELECT CASE ENDSELECT IS

%token WEND DO LOOP UNTIL WHILE EXITDO
%token DOWHILE DOUNTIL LOOPWHILE LOOPUNTIL

%token SUB ENDSUB EXITSUB DECLARESUB
%token FUNCTION ENDFUNCTION EXITFUNCTION DECLAREFUNCTION
%token CALL

%token FOR TO STEP NEXT EXITFOR
%token GOTO GOSUB RETURN
%token END
%token LET

%token NUMBER STRING
%token IDENTIFIER FUNCTIONNAME SUBNAME CONSTNAME

%token PRINT USING FILENUM LPRINT

%token DEFINT DEFSNG DEFDBL DEFSTR DEFLNG
%token INTEGER SINGLE DOUBLE STRING LONG AS
%token TYPE ENDTYPE
%token DIM SHARED CONST

%%
/* Grammar rules */

program:
		definitionlist
;

definitionlist:
		definitionlist definition
	|	/* empty */
;

definition:
		sub
	|	function
	|	declaresub
	|	declarefunction
	|	typedef
	|	deftypedef
	|	constdef
	|	longstatement
	|	CRLF /* empty line is ok in main program */
;

functionparamlistdeclaration:
		'(' paramlistdeclaration ')'
	|	'(' ')'
	|	/* empty */
;
sub:
		SUB SUBNAME functionparamlistdeclaration CRLF longstatements ENDSUB CRLF
;
function:
		FUNCTION FUNCTIONNAME functionparamlistdeclaration CRLF longstatements ENDFUNCTION CRLF
;
declaresub:
		DECLARESUB newsubname functionparamlistdeclaration CRLF
;
declarefunction:
		DECLAREFUNCTION newfunctionname functionparamlistdeclaration CRLF
;
paramlistdeclaration:
		paramlistdeclaration ',' paramdeclaration
	|	paramdeclaration
;
paramdeclaration:
		IDENTIFIER AS type
	|	IDENTIFIER
;


typedef:
		TYPE IDENTIFIER CRLF typeelements ENDTYPE CRLF
;
typeelements:
		typeelements typeelement
	|	typeelement
;
typeelement:
		typeelementdef CRLF
	|	CRLF /* empty line is ok in typedef */
;
typeelementdef:
		typeelementdef ',' IDENTIFIER AS type /* extension */
	|	IDENTIFIER AS type
;

deftypedef:
		DEFINT characterranges CRLF
	|	DEFSNG characterranges CRLF
	|	DEFDBL characterranges CRLF
	|	DEFSTR characterranges CRLF
	|	DEFLNG characterranges CRLF
;
characterranges:
		characterranges ',' characterrange
	|	characterrange
;
characterrange:
		onechar '-' onechar
	|	onechar
;

constdef:
		CONST IDENTIFIER '=' expr CRLF
;

/* This does not terminate with anything */
shortstatementonly:
		simplestatement
	|	shortif
	|	shortwhile
	|	shortdoloop
	|	shortfornext
	|	shortselect
;
shortstatementsnotempty:
		shortstatementsnotempty COLONCHAR shortstatementonly
	|	shortstatementonly
;
shortstatements:
		shortstatementsnotempty
	|	/* empty */
;

shortif:
		IF boolexpr THEN shortifblock ELSE shortifblock
	|	IF boolexpr THEN              ELSE shortifblock
	|	IF boolexpr THEN shortifblock ELSE 
	|	IF boolexpr THEN shortifblock
;
shortifblock:
		NUMBER
	|	shortstatementsnotempty
;
shortdoloop:
		DO COLONCHAR shortstatements LOOP
	|	DOWHILE boolexpr COLONCHAR shortstatements LOOP
	|	DOUNTIL boolexpr COLONCHAR shortstatements LOOP
	|	DO COLONCHAR shortstatements LOOPWHILE boolexpr
	|	DO COLONCHAR shortstatements LOOPUNTIL boolexpr
;
shortwhile:
		WHILE boolexpr COLONCHAR shortstatements WEND
;
shortfornext:
		FOR target '=' numexpr TO numexpr forstep COLONCHAR shortstatements NEXT
;
shortselect:
		SELECT CASE expr COLONCHAR shortcaselist ENDSELECT
;
shortcaselist:
		shortcaselist CASE casevalues COLONCHAR shortstatements
	|	/* empty */
;

longstatement:
		longstructure separatorsnotempty
	|	simplestatement separatorsnotempty
	|	shortif separatorsnotempty
;
longstructure:
		longif
	|	longwhile
	|	longdoloop
	|	longfornext
	|	longselect
;
longif:
		IF boolexpr THEN CRLF longstatements longelselist ENDIF
;
longelselist:
		ELSEIF boolexpr THEN separatorsnotempty longstatements longelselist
	|	ELSE separatorsnotempty longstatements
	|	/* empty */
;
longdoloop:
		DO separatorsnotempty longstatements LOOP
	|	DOWHILE boolexpr separatorsnotempty longstatements LOOP
	|	DOUNTIL boolexpr separatorsnotempty longstatements LOOP
	|	DO separatorsnotempty longstatements LOOPWHILE boolexpr
	|	DO separatorsnotempty longstatements LOOPUNTIL boolexpr
;
longwhile:
		WHILE boolexpr separatorsnotempty longstatements WEND
;
longfornext:
		FOR target '=' numexpr TO numexpr forstep separatorsnotempty longstatements NEXT
;
longselect:
		SELECT CASE expr separatorsnotempty longcaselist ENDSELECT
;
longcaselist:
		longcaselist CASE casevalues separatorsnotempty longstatements
	|	/* empty */
;

longstatements:
		longstatementsnotempty
	|	/* empty */
;
longstatementsnotempty:
		longstatementsnotempty longstatement
	|	longstatement
;

forstep:
		STEP numexpr
	|	/* empty */
;
casevalue:
		ELSE
	|	IS '>' expr
	|	IS '<' expr
	|	IS '=' expr
	|	IS LESSOREQUAL expr
	|	IS GREATEROREQUAL expr
	|	IS NOTEQUAL expr
	|	expr TO expr
	|	expr
;
casevalues:
		casevalues ',' casevalue
	|	casevalue
;

separatorsnotempty:
		separatorsnotempty separator
	|	separator
;
separator:
		COLONCHAR
	|	CRLF
;

anylabel:
		NUMBER
	|	label COLONCHAR
;


/* Something that is not a control structure */
/* This does not terminate with anything */
simplestatement:
		LET assignment
	|	assignment
	|	GOTO label
	|	GOSUB label
	|	GOTO NUMBER
	|	GOSUB NUMBER
	|	RETURN
	|	EXITDO
	|	EXITFOR
	|	EXITSUB
	|	EXITFUNCTION
	|	END
	|	subcall		/* This causes 1 shift/reduce conflict */
	|	print		/* This causes 6 shift/reduce conflicts */
	|	dim
;

assignment:
		target '=' expr
;

subcall:
		SUBNAME subcallparamlist
	|	CALL SUBNAME '(' subcallparamlist ')'
	|	CALL SUBNAME
;
subcallparamlist:
		subcallparams
	|	/* empty */
;
subcallparams:
		subcallparams ',' expr
	|	expr
;

print:
		PRINT FILENUM ',' printspec
	|	PRINT printspec
	|	LPRINT printspec
;
printspec:
		USING stringexpr ';' printparams
	|	printparams
;
printparams:
		printparams ';' expr  /* list; expr */
	|	printparams ';'       /* list;  */
	|	printparams ',' expr  /* list, expr */
	|	printparams ','       /* list,  */
	|	expr                  /* expr */
	|	/* empty */
;

dim:
		DIM SHARED dimdefs
	|	DIM dimdefs
;
dimdefs:
		dimdefs ',' dimdef
	|	dimdef
;
dimdef:
		IDENTIFIER '(' dimindexdefs ')' AS type
	|	IDENTIFIER '(' dimindexdefs ')'
	|	IDENTIFIER AS type
	|	IDENTIFIER
;
dimindexdefs:
		dimindexdefs ',' dimindexdef
	|	dimindexdef
;
dimindexdef:
		numexpr TO numexpr
	|	numexpr
;

numexpr:
		expr	/* FIXME - check expression type */
;
boolexpr:
		expr	/* FIXME - check expression type */
;
stringexpr:
		expr	/* FIXME - check expression type */
;

expr:
		expr IMP expr
	|	expr EQV expr
	|	expr XOR expr
	|	expr OR expr
	|	expr AND expr
	|	NOT expr
	
	|	expr '>' expr
	|	expr '<' expr
	|	expr '=' expr
	|	expr LESSOREQUAL expr
	|	expr GREATEROREQUAL expr
	|	expr NOTEQUAL expr

	|	expr '+' expr
	|	expr '-' expr
	|	expr MOD expr
	|	expr '\\' expr
	|	expr '/' expr
	|	expr '*' expr
	|	expr '^'expr
	
	|       '-' expr %prec NEGATIONPRIORITY

	|	'(' expr ')'
	|	var
	|	functioncall
	|	NUMBER
	|	STRING
	|	CONSTNAME
;

onechar:
		IDENTIFIER /* FIXME: add length check */
;
label:
		IDENTIFIER
;

functioncall:
		FUNCTIONNAME '(' subcallparams ')'
	|	FUNCTIONNAME
;

indexlist:
		indexlist ',' numexpr
	|	numexpr
;
var:
		IDENTIFIER '(' indexlist ')'
	|	IDENTIFIER
;

target:
		var
;

type:
		INTEGER
	|	SINGLE
	|	DOUBLE
	|	STRING
	|	LONG
	|	usertype
;
usertype:
		IDENTIFIER
;
newsubname:
		IDENTIFIER
		{/* fprintf(stderr, "Got new sub name: '%s'\n", yytext); */}
;
newfunctionname:
		IDENTIFIER
;

	

%%
/* Additional C code */


int main(void)
{
    yydebug=1;
    yyparse();
    return 0;
}
